#ifndef _MERGEOS_PROTO_H_
#define _MERGEOS_PROTO_H_

// graphics/palette.c
PUBLIC void init_palette();
PUBLIC void set_palette(int start, int end, u8 *rgb);

// graphics/text.c
PUBLIC void put_char(char c);
PUBLIC void put_str(const char *str);
PUBLIC void put_int(int input);

// kernel/gdt.c
PUBLIC void init_gdt();
PUBLIC u32 seg2phys(u16 seg);
PUBLIC void init_descriptors();
PUBLIC void init_descriptor(DESCRIPTOR *p_desc, u32 base, u32 limit, u16 attribute);

// kernel/intr.c
PUBLIC void init_idt();
PUBLIC void exception_handler(int vec_no, int error_code, int eip, int cs, int eflags);

// kernel/i8259.c
PUBLIC void init_8259A();
PUBLIC void spurious_irq(int irq);
PUBLIC void put_irq_handler(int irq, irq_handler handler);

// kernel/binfo.c
PUBLIC void init_vram();

// kernel/clock.c
PUBLIC void clock_handler(int irq);
PUBLIC void milli_delay(int ms);
PUBLIC void init_clock();

// kernel/fifo.c
PUBLIC void fifo_init(FIFO32 *fifo, int size, u32 *buf);
PUBLIC int fifo_put(FIFO32 *fifo, u32 data);
PUBLIC u32 fifo_get(FIFO32 *fifo);
PUBLIC int fifo_status(FIFO32 *fifo);

// kernel/memory.c
PUBLIC u32 memtest(u32 start, u32 end);
PUBLIC u32 memtest_sub(u32 start, u32 end);

PUBLIC void memman_init(MEMMAN *man);
PUBLIC u32 memman_total(MEMMAN *man);
PUBLIC u32 memman_alloc(MEMMAN *man, u32 size);
PUBLIC int memman_free(MEMMAN *man, u32 addr, u32 size);

PUBLIC u32 memman_alloc_4k(MEMMAN *man, u32 size);
PUBLIC int memman_free_4k(MEMMAN *man, u32 addr, u32 size);

PUBLIC void *kmalloc(u32 size);
PUBLIC void kfree(void *addr, u32 size);

PUBLIC void init_mem();

// kernel/kernel.asm
PUBLIC void restart();

// device/keyboard/keyboard.c
PUBLIC void keyboard_handler(int irq);
PUBLIC void init_keyboard();
PUBLIC void keyboard_read();

// device/keyboard/keydrive.c
PUBLIC void keyboard_driver();
PUBLIC void in_process(u32 key);

// kernel/proc.c
PUBLIC void schedule();

// kernel/process.c
PUBLIC void init_proc();
PUBLIC void process_start(char *name, task_f proc, int prio, PROC_MODE mode);

// lib/kliba.asm
PUBLIC u32 load_eflags();
PUBLIC void store_eflags(u32 eflags);
PUBLIC void out_byte(u16 port, u8 value);
PUBLIC u8 in_byte(u16 port);
PUBLIC void disable_int();
PUBLIC void enable_int();
PUBLIC u32 load_cr0();
PUBLIC void store_cr0(u32 cr0);
PUBLIC void disable_irq(int irq);
PUBLIC void enable_irq(int irq);
PUBLIC void load_gdtr(u8 *gdt_ptr);

// lib/klib.c
PUBLIC char *itoa(char *str, int num);
PUBLIC void delay(int time);

// lib/assert.c
PUBLIC void panic_spin(char *filename, int line, const char *func, const char *condition);

// lib/istatus.c
PUBLIC enum intr_status intr_enable();
PUBLIC enum intr_status intr_disable();
PUBLIC enum intr_status intr_get_status();
PUBLIC enum intr_status intr_set_status(enum intr_status status);

// system calls

// kernel/proc.c
PUBLIC int sys_get_ticks();

// kernel/kernel.asm
PUBLIC void sys_call();

// kernel/syscall.asm
PUBLIC int get_ticks();

#endif