#include "include.h"

PUBLIC void init_proc()
{
    procctl = (PROCCTL *) kmalloc(sizeof(PROCCTL));
    memset(procctl, 0, sizeof(procctl));
}

PUBLIC void process_start(char *name, task_f proc, int prio, PROC_MODE mode)
{
    PROCESS *p_proc;
    char    *p_task_stack = kmalloc(0x8000);
    u16      selector_ldt = SELECTOR_LDT_FIRST + procctl->running;
    int      i, rpl, priv, eflags;
    for (i = 0; i < MAX_TASKS; i++) {
        if (procctl->tasks0[i].flags == 0) {
            p_proc = &procctl->tasks0[i];
            break;
        }
    }

    if (mode == KERNEL) {
        rpl = RPL_TASK;
        priv = PRIVILEGE_TASK;
        eflags = 0x1202;
    } else {
        rpl = RPL_USER;
        priv = PRIVILEGE_USER;
        eflags = 0x202;
    }

    strcpy(p_proc->p_name, name);
    p_proc->pid = i;

    p_proc->ldt_sel = selector_ldt;

    memcpy(&p_proc->ldts[0], &gdt[SELECTOR_KERNEL_CS >> 3], sizeof(DESCRIPTOR));
    p_proc->ldts[0].attr1 = DA_C | priv << 5;
    memcpy(&p_proc->ldts[1], &gdt[SELECTOR_KERNEL_DS >> 3], sizeof(DESCRIPTOR));
    p_proc->ldts[1].attr1 = DA_DRW | priv << 5;

    p_proc->regs.cs = ((0 * 8) & SA_RPL_MASK & SA_TI_MASK) | SA_TIL | rpl;
    p_proc->regs.ds = ((1 * 8) & SA_RPL_MASK & SA_TI_MASK) | SA_TIL | rpl;
    p_proc->regs.es = ((1 * 8) & SA_RPL_MASK & SA_TI_MASK) | SA_TIL | rpl;
    p_proc->regs.fs = ((1 * 8) & SA_RPL_MASK & SA_TI_MASK) | SA_TIL | rpl;
    p_proc->regs.ss = ((1 * 8) & SA_RPL_MASK & SA_TI_MASK) | SA_TIL | rpl;
    p_proc->regs.gs = (SELECTOR_KERNEL_GS & SA_RPL_MASK) | rpl;
    p_proc->regs.eip = (u32) proc;
    p_proc->regs.esp = (u32) p_task_stack;
    p_proc->regs.eflags = eflags;

    p_proc->ticks = p_proc->priority = prio;

    procctl->tasks0[i].flags = 1;
    procctl->running++;

    init_descriptor(&gdt[selector_ldt >> 3],
            vir2phys(seg2phys(SELECTOR_KERNEL_DS), p_proc->ldts),
            LDT_SIZE * sizeof(DESCRIPTOR) - 1,
            DA_LDT);

    u16 *p_gdt_limit = (u16 *) (&gdt_ptr[0]);
    u32 *p_gdt_base  = (u32 *) (&gdt_ptr[2]);
    *p_gdt_limit = GDT_SIZE * sizeof(DESCRIPTOR) - 1;
    *p_gdt_base  = (u32) &gdt;

    load_gdtr(gdt_ptr);
}