%include "sconst.inc"

[section .text]

global disable_int, enable_int
global out_byte, in_byte
global load_eflags, store_eflags
global load_cr0, store_cr0
global load_gdtr
global disable_irq, enable_irq

disable_int:
    cli
    ret

enable_int:
    sti
    ret

out_byte:
    mov edx, [esp + 4]
    mov al, [esp + 8]
    out dx, al
    nop
    nop
    ret

in_byte:
    mov edx, [esp + 4]
    xor eax, eax
    in al, dx
    nop
    nop
    ret

load_eflags:
    pushfd
    pop eax
    ret

store_eflags:
    mov eax, [esp + 4]
    push eax
    popfd
    ret

load_cr0:
    mov eax, cr0
    ret

store_cr0:
    mov eax, [esp + 4]
    mov cr0, eax
    ret

load_gdtr:
    mov eax, [esp + 4]
    lgdt [eax]
    ret

disable_irq:
    mov ecx, [esp + 4]
    pushf
    cli
    mov ah, 1
    rol ah, cl
    cmp cl, 8
    jae disable_8
disable_0:
    in al, INT_M_CTLMASK
    test al, ah
    jnz dis_already
    or al, ah
    out INT_M_CTLMASK, al
    popf
    mov eax, 1
    ret
disable_8:
    in al, INT_S_CTLMASK
    test al, ah
    jnz dis_already
    or al, ah
    out INT_S_CTLMASK, al
    popf
    mov eax, 1
    ret
dis_already:
    popf
    xor eax, eax
    ret

enable_irq:
    mov ecx, [esp + 4]
    pushf
    cli
    mov ah, ~1
    rol ah, cl
    cmp cl, 8
    jae enable_8
enable_0:
    in al, INT_M_CTLMASK
    and al, ah
    out INT_M_CTLMASK, al
    popf
    ret
enable_8:
    in al, INT_S_CTLMASK
    and al, ah
    out INT_S_CTLMASK, al
    popf
    ret
